/*
Copyright (C) 2006  The PARI group.

This file is part of the PARI package.

PARI/GP is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation. It is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY WHATSOEVER.

Check the License for details. You should have received a copy of it, along
with the package; see the file 'COPYING'. If not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */

#include "pari.h"
#include "paripriv.h"
#include "anal.h"
#include "tree.h"
#include "opcode.h"

/***************************************************************************
 **                                                                       **
 **                           Byte-code compiler                          **
 **                                                                       **
 ***************************************************************************/

static THREAD gp2c_stack s_opcode, s_data;
static THREAD long *opcode;
static THREAD GEN *data;
static THREAD long offset=-1;

void
pari_init_compiler(void)
{
  stack_init(&s_opcode,sizeof(*opcode),(void **)&opcode);
  stack_init(&s_data,sizeof(*data),(void **)&data);
}

static void
getcodepos(long *sop, long *sdat, long *off)
{
  *sop=s_opcode.n;
  *sdat=s_data.n;
  *off=offset;
  offset=s_data.n-1;
}

static GEN
getclosure(long sop, long sdat, long off)
{
  long lop =s_opcode.n+1-sop;
  long ldat=s_data.n+1-sdat;
  GEN cl=cgetg(3,t_VEC);
  long i;
  gel(cl,1) = cgetg(lop,  t_VECSMALL);
  gel(cl,2) = cgetg(ldat, t_VEC);
  for(i=1;i<lop;i++)
    mael(cl, 1, i) = opcode[i+sop-1];
  s_opcode.n=sop;
  for(i=1;i<ldat;i++)
  {
    gmael(cl, 2, i) = gcopy(data[i+sdat-1]);
    gunclone(data[i+sdat-1]);
  }
  s_data.n=sdat;
  offset=off;
  return cl;
}

static void
op_push(op_code o, long x)
{
  long n=stack_new(&s_opcode);
  long m=stack_new(&s_opcode);
  opcode[n]=o;
  opcode[m]=x;
}

static long
data_push(GEN x)
{
  long n=stack_new(&s_data);
  data[n] = gclone(x);
  return n-offset;
}
 
static void compilenode(long n, int mode);

typedef enum {PPend,PPstd,PPdefault,PPdefaultmulti,PPstar,PPauto} PPproto;

static PPproto
parseproto(char const **q, char *c)
{
  char  const *p=*q;
  long i;
  switch(*p)
  {
  case 0:
  case '\n':
    return PPend;
  case 'D':
    switch(p[1])
    {
    case 0:
      pari_err(talker,"incomplete prototype");
    case 'G':
    case '&':
    case 'V':
    case 'I':
    case 'E':
    case 'n':
      *c=p[1];
      *q=p+2;
      return PPdefault;
    default:
      for(i=0;*p && i<2;p++) i+=*p==',';
      if (i<2)
        pari_err(talker,"incomplete prototype");
      *c=p[-2];
      *q=p;
      return PPdefaultmulti;
    }
    break;
  case 'p':
  case 'P':
  case 'f':
    *c=*p;
    *q=p+1;
    return PPauto;
  case '&':
    *c='*';
    *q=p+1;
    return PPstd;
  case 'V':
    if (p[1]=='=')
    {
      *c=*p;
      p+=2;
      *q=p+1;
      return PPstar;
    }
  case 's': /*fall through*/
    if (p[1]=='*')
    {
      *c=*p++;
      *q=p+1;
      return PPstar;
    }
  default:/*fall through*/
    *c=*p;
    *q=p+1;
    return PPstd;
  }
}
/*supported types:
 * type: Gsmall, Ggen, Gvar, Gvoid, Gvec
 * mode: Gsmall, Ggen, Gvar, Gvoid
 */
static void
compilecast(long n, int type, int mode)
{
  if (type==mode) return;
  switch (mode)
  {
  case Gsmall:
    if (type==Ggen)        op_push(OCitos,0);
    else if (type==Gvoid)  op_push(OCpushlong,0);
    else pari_err(talker2,"this should be an integer",
             tree[n].start, get_origin());
    break;
  case Ggen:
    if (type==Gsmall)      op_push(OCstoi,0);
    else if (type==Gvoid)  op_push(OCpushlong,(long)gnil);
    else if (type==Gvar )  op_push(OCpolx,0);
    break;
  case Gvoid:
    op_push(OCpop, 1);
    break;
  case Gvar:
    if (type==Ggen)        op_push(OCvarn,0);
    else pari_err(varer1, tree[n].start, get_origin());
     break;
  default:
    pari_err(bugparier,"compilecast, type unknown %ld",mode);
  }
}

static char *
entryname(long x)
{
  return GSTR(value[x].gen);
}

static entree *
getentry(long n)
{
  long x=tree[n].x;
  char *name=entryname(x);
  return fetch_entry(name);
}

static entree *
getfunc(long n)
{
  return do_alias(getentry(n));
}

static GEN 
listtogen(long n, long f)
{
  GEN z;
  long x,i,nb;
  if (n==-1 || tree[n].f==Fnoarg) return cgetg(1,t_VECSMALL);
  for(x=n, i=0; tree[x].f==f ;x=tree[x].x, i++);
  nb=i+1;
  z=cgetg(nb+1, t_VECSMALL);
  for (x=n; i>0; z[i+1]=tree[x].y, x=tree[x].x, i--);
  z[1]=x;
  return z;
}

static entree *
getlvalue(long n)
{ 
  while (tree[n].f==Ffacteurmat)
    n=tree[n].x;
  return getentry(n);
}

static void
compilelvalue(long n)
{ 
  if (tree[n].f==Fentry)
    return;
  else
  {
    long x=tree[n].x;
    long y=tree[n].y;
    long yx=tree[y].x;
    long yy=tree[y].y;
    long f=tree[y].f;
    if (tree[x].f==Ffacteurmat && f==Fmatrix && yy==-1 && 
        tree[tree[x].y].f==FmatrixL)
    {
      compilelvalue(tree[x].x);
      compilenode(tree[tree[x].y].x,Gsmall);
      compilenode(yx,Gsmall);
      op_push(OCcompo2ptr,0);
      return;
    }
    compilelvalue(x);
    compilenode(yx,Gsmall);
    if (f==Fmatrix && yy==-1)
      op_push(OCcompo1ptr,0);
    else
    {
      switch(f)
      {
      case Fmatrix:
        compilenode(yy,Gsmall);
        op_push(OCcompo2ptr,0);
        break;
      case FmatrixR:
        op_push(OCcompoCptr,0);
        break;
      case FmatrixL:
        op_push(OCcompoLptr,0);
        break;
      }
    }
  } 
}

static void 
compilefacteurmat(long n, int mode)
{ 
  long x=tree[n].x;
  long y=tree[n].y;
  long yx=tree[y].x;
  long yy=tree[y].y;
  long f=tree[y].f;
  compilenode(x,Ggen);
  compilenode(yx,Gsmall);
  if (f==Fmatrix && yy==-1)
  {
    op_push(OCcompo1,mode);
    return;
  }
  switch(f)
  {
  case Fmatrix:
    compilenode(yy,Gsmall);
    op_push(OCcompo2,mode);
    break;
  case FmatrixR:
    op_push(OCcompoC,0);
    compilecast(n,Gvec,mode);
    return;
  case FmatrixL:
    op_push(OCcompoL,0);
    compilecast(n,Gvec,mode);
    return;
  default:
    pari_err(bugparier,"compilefacteurmat");
  }
}

static void
compilevec(long n, long mode, op_code op)
{
  pari_sp ltop=avma;
  long x=tree[n].x;
  long i;
  GEN arg=listtogen(x,Fmatrixelts);
  long l=lg(arg);
  op_push(op,l);
  for(i=1;i<l;i++)
  {
    compilenode(arg[i],Ggen);
    op_push(OCstackgen,i);
  }
  avma=ltop;
  compilecast(n,Gvec,mode);
}

static void
compilemat(long n, long mode)
{
  pari_sp ltop=avma;
  long x=tree[n].x;
  long i,j;
  GEN line=listtogen(x,Fmatrixlines);
  long lglin = lg(line), lgcol=0;
  op_push(OCpushlong, lglin);
  if (lglin==1)
    op_push(OCmat,1);
  for(i=1;i<lglin;i++)
  {
    GEN col=listtogen(line[i],Fmatrixelts);
    long l=lg(col), k;
    if (i==1)
    {
      lgcol=l;
      op_push(OCmat,lgcol);
    }
    else if (l!=lgcol)
      pari_err(talker2,"matrix must be rectangular",
          tree[line[i]].start,get_origin());
    k=i;
    for(j=1;j<lgcol;j++)
    {
      k-=lglin;
      compilenode(col[j], Ggen);
      op_push(OCstackgen,k);
    }
  }
  avma=ltop;
  compilecast(n,Gvec,mode);
}


static GEN
cattovec(long n, long fnum)
{
  long x=n, i=0, nb;
  GEN stack;
  if (tree[n].f==Fnoarg) return cgetg(1,t_VECSMALL);
  while(1)
  {
    long xx=tree[x].x;
    long xy=tree[x].y;
    if (tree[x].f!=Ffunction || xx!=fnum) break;
    x=tree[xy].x;
    i++;
  }
  nb=i+1;
  stack=cgetg(nb+1,t_VECSMALL);
  for(x=n;i>0;i--)
  {
    long y=tree[x].y;
    x=tree[y].x;
    stack[i+1]=tree[y].y;
  }
  stack[1]=x;
  return stack;
}

static void 
compilefunc(long n, int mode)
{
  pari_sp ltop=avma;
  long i,j;
  long x=tree[n].x;
  long y=tree[n].y;
  char *name=entryname(x);
  long ret;
  char const *p,*q;
  char c;
  PPproto mod;
  GEN arg=listtogen(y,Flistarg);
  long nbpointers=0;
  long nb=lg(arg)-1;
  entree *ep = getfunc(n);
  if (EpVALENCE(ep)==EpVAR || EpVALENCE(ep)==EpGVAR)
    pari_err(talker2,"not a function in function call",
        tree[n].start, get_origin());
  if (EpVALENCE(ep)==EpUSER|| EpVALENCE(ep)==EpNEW)
  {
    for (j=1;j<=nb;j++)
      if (tree[arg[j]].f!=Fnoarg)
        compilenode(arg[j], Ggen);
      else
        op_push(OCpushlong,0);
    op_push(OCpushlong, nb);
    if (tree[n].f==Fderfunc)
      op_push(OCderivuser, (long) ep);
    else
      op_push(OCcalluser, (long) ep);
    compilecast(n,Ggen,mode);
    avma=ltop;
    return;
  }
  if (!strcmp(name,"if") && mode==Gvoid)
    ep=is_entry("_void_if");
  if (!strcmp(name,"local"))
  {
    if (tree[n].f==Fderfunc)
      pari_err(talker2,"can't derive this",tree[n].start,get_origin());
    compilecast(n,Gvoid,mode);
    avma=ltop;
    return;
  }
  if (!strcmp(name,"global"))
  {
    if (tree[n].f==Fderfunc)
      pari_err(talker2,"can't derive this",tree[n].start,get_origin());
    for (i=1;i<lg(arg);i++)
    {
      long a=arg[i];
      long en;
      if (tree[a].f==Faffect)
      {
        compilenode(tree[a].y,Ggen);
        a=tree[a].x;
      }
      else
        op_push(OCpushlong,0);
      if (tree[a].f==Ftag)
        a=tree[a].x;
      en=(long)getentry(a);
      op_push(OCglobalvar,en);
    }
    compilecast(n,Gvoid,mode);
    avma=ltop;
    return;
  }
  if (!strcmp(name,"O") || (compatible==OLDALL && !strcmp(name,"o")))
  {
    long a=tree[n].y;
    if (tree[n].f==Fderfunc)
      pari_err(talker2,"can't derive this",tree[n].start,get_origin());
    if (tree[a].f!=Ffunction || tree[a].x!=OPpow)
    {
      compilenode(a,Ggen);
      op_push(OCpushlong,1);
    }
    else
    {
      long y=tree[a].y;
      compilenode(tree[y].x,Ggen);
      compilenode(tree[y].y,Gsmall);
    }
    op_push(OCcallgen,(long)ep);
    compilecast(n,Ggen,mode);
    avma=ltop;
    return;
  }
  if (x==OPtrans && tree[y].f==Fvec)
  {
    avma=ltop;
    compilevec(y, mode, OCcol);
    return;
  }
  if (x==OPcat)
    pari_err(talker2,"expected character: ',' or ')' instead of",
        tree[arg[1]].end, get_origin());
  p=ep->code;
  if (!ep->value)
    pari_err(talker2,"unknown function",tree[n].start, get_origin());
  if (*p == 'v') { ret = RET_VOID; p++; }
  else if (*p == 'i') { ret = RET_INT;  p++; }
  else if (*p == 'l') { ret = RET_LONG; p++; }
  else ret = RET_GEN;
  if (tree[n].f==Fderfunc && (ret!=RET_GEN || *p!='G'))
    pari_err(talker2,"can't derive this",tree[n].start,get_origin());
  i=0; j=1;
  if (*p)
  {
    q=p; 
    while((mod=parseproto(&p,&c))!=PPend)
    {
      if (j<=nb && tree[arg[j]].f!=Fnoarg 
                && (mod==PPdefault || mod==PPdefaultmulti))
        mod=PPstd;
      switch(mod)
      {
      case PPstd:
        if (j>nb)
          pari_err(talker2,"too few arguments",
              tree[n].end-1, get_origin());
        if (tree[arg[j]].f==Fnoarg && c!='I' && c!='E')
          pari_err(talker2,"missing mandatory argument",
              tree[arg[j]].end, get_origin());
        switch(c)
        {
        case 'G':
          compilenode(arg[j++],Ggen);
          break;
        case 'M':
        case 'L':
          compilenode(arg[j++],Gsmall);
          break;
        case 'n':
          compilenode(arg[j++],Gvar);
          break;
        case '&': case '*': 
          {
            long a=arg[j++];
            entree *ep;
            if (c=='&')
            {
              if (tree[a].f!=Frefarg)
                pari_err(talker2,"expected character: '&'",
                    tree[a].start, get_origin());
              a=tree[a].x;
            }
            ep=getlvalue(a);
            op_push(OCnewptr, (long) ep);
            compilelvalue(a);
            op_push(OCpushptr, 0);
            nbpointers++;
            break;
          }
        case 'I':
        case 'E':
          {
            long sop,sdat,off;
            long a=arg[j++];
            int type=c=='I'?Gvoid:Ggen;
            getcodepos(&sop,&sdat,&off);
            if (tree[a].f==Fnoarg)
              compilecast(a,Gvoid,type);
            else
              compilenode(a,type);
            op_push(OCpushgen, data_push(getclosure(sop,sdat,off)));
            break;
          }
        case 'V':
        case 'S':
          {
            entree *ep = getentry(arg[j++]);
            op_push(OCpushlong, (long)ep);
            break;
          }
        case 'r':
          {
            long a=arg[j++];
            if (tree[a].f==Fentry)
            {
              op_push(OCpushgen, data_push(value[tree[a].x].gen));
              op_push(OCtostr, 1);
            }
            else
            {
              compilenode(a,Ggen);
              op_push(OCtostr, 1);
            }
            break;
          }
        case 's':
          {
            GEN g = cattovec(arg[j++], OPcat);
            long l, nb = lg(g)-1;
            for(l=1; l<=nb; l++)
              compilenode(g[l], Ggen);
            op_push(OCtostr, nb);
            break;
          }
        default:
          pari_err(talker,"Unknown prototype code `%c' for `%s'",c,name);
        }
        break;
      case PPauto:
        switch(c)
        {
        case 'p':
          op_push(OCprecreal,0);
          break;
        case 'P':
          op_push(OCprecdl,0);
          break;
        case 'f':
          {
            static long foo;
            op_push(OCpushlong,(long)&foo);
            break;
          }
        }
        break;
      case PPdefault:
        j++;
        switch(c)
        {
        case 'G':
        case '&':
        case 'V':
        case 'r':
        case 'E':
        case 'I':
          op_push(OCpushlong,0);
          break;
        case 'n':
          op_push(OCpushlong,-1);
          break;
        default:
          pari_err(talker,"Unknown default in prototype code `%c' for `%s'",c,name);
        }
        break;
      case PPdefaultmulti:
        j++;
        switch(c)
        {
        case 'G':
          op_push(OCpushgen,data_push(strntoGENstr(q+1,p-4-q)));
          op_push(OCcallgen,(long)is_entry("eval"));
          break;
        case 'L':
        case 'M':
          op_push(OCpushlong,strtol(q+1,NULL,10));
          break;
        case 'r':
        case 's':
          if (q[1]=='"' && q[2]=='"')
            op_push(OCpushlong,(long)"");
          else
            pari_err(impl,"prototype not supported");
          break;
        default:
          pari_err(talker,"Unknown default in prototype code `%c' for `%s'",c,name);
        }
        break;
      case PPstar:
        switch(c)
        {
        case 'V':
          {
            long x=tree[arg[j]].x;
            long y=tree[arg[j]].y;
            entree *ep;
            if (tree[arg[j]].f!=Faffect)
              pari_err(talker2,"expected character: '=' instead of",
                  tree[arg[j]].end, get_origin());
            ep = getentry(x);
            op_push(OCpushlong, (long)ep);
            compilenode(y,Ggen);
            i++; j++;
          }
          break;
        case 's':
          {
            long n=nb+1-j;
            long k,l,l1,m;
            GEN g=cgetg(n+1,t_VEC);
            for(l1=0,k=1;k<=n;k++)
            {
              gel(g,k)=cattovec(arg[j+k-1],OPcat);
              l1+=lg(g[k])-1;
            }
            op_push(OCvec, l1+1);
            for(m=1,k=1;k<=n;k++)
              for(l=1;l<lg(g[k]);l++,m++)
              {
                compilenode(mael(g,k,l),Ggen);
                op_push(OCstackgen,m); 
              }
            j=nb+1;
            break;
          }
        default:
          pari_err(talker,"Unknown prototype code `%c*' for `%s'",c,name);
        }
        break;
      default:
        pari_err(bugparier,"PPproto %d in gencallfunc",mod);
      }
      i++;
      q=p;
    }
  }
  if (j<=nb)
    pari_err(talker2,"too many arguments",tree[arg[j]].start,get_origin());
  switch (ret)
  {
  case RET_GEN: 
    if (tree[n].f==Fderfunc)
      op_push(OCderivgen, (long) ep);
    else
      op_push(OCcallgen, (long) ep);
    compilecast(n,Ggen,mode);
    break;
  case RET_INT:
    op_push(OCcallint, (long) ep);
    compilecast(n,Gsmall,mode);
    break;
  case RET_LONG:
    op_push(OCcalllong, (long) ep);
    compilecast(n,Gsmall,mode);
    break;
  case RET_VOID:
    op_push(OCcallvoid, (long) ep);
    compilecast(n,Gvoid,mode);
    break;
  }
  if (nbpointers) op_push(OCendptr,nbpointers);
  avma=ltop;
}

static void
compilenode(long n, int mode)
{
  long x,y;
  if (n<0)
    pari_err(bugparier,"compilenode");
  x=tree[n].x;
  y=tree[n].y;
   
  switch(tree[n].f)
  {
  case Fseq:
    if (tree[x].f!=Fnoarg)
      compilenode(x,Gvoid);
    compilenode(y,mode);
    return;
  case Ffacteurmat:
    compilefacteurmat(n,mode);
    break;
  case Faffect:
    {
      struct node_loc l;
      l.start=tree[n].start;
      l.end=tree[n].end;
      compilefunc(newopcall(OPstore,x,y,&l),mode);
    }
    break;
  case Fconst:
    if (value[x].type!=CSTquote)
    {
      if (mode==Gvoid) return;
      if (mode==Gvar)
        pari_err(varer1,tree[n].start,get_origin());
    }
    if (mode==Gsmall)
      pari_err(talker2,"this should be a small integer",
          tree[n].start,get_origin());
    switch(value[x].type)
    {
    case CSTreal: 
      op_push(OCpushreal, data_push(value[x].gen));
      return;
    case CSTgen: 
      op_push(OCpushgen,  data_push(value[x].gen));
      return;
    case CSTstr:
      op_push(OCpushgen,  data_push(value[x].gen));
      return;
    case CSTquote:
      op_push(OCpushlong, fetch_user_var(entryname(x)));
      compilecast(n,Gvar, mode);
      return;
    }
    break;
  case Fsmall:
    if (mode==Ggen)
      op_push(OCpushstoi, x);
    else
    {
      op_push(OCpushlong, x);
      compilecast(n,Gsmall,mode);
    }
    return;
  case Fvec:
    compilevec(n, mode, OCvec);
    return;
  case Fmat:
    compilemat(n, mode);
    return;
  case Frefarg:
    pari_err(talker,"unexpected &"); 
    break;
  case Fentry:
    {
      entree *ep=getentry(n);
      if (!EpSTATIC(ep))
      {
        op_push(OCpushentryval,(long)ep);
        compilecast(n,Ggen,mode);
        break;
      }
    }
  case Fderfunc: /*Fall through*/
  case Ffunction:
    compilefunc(n, mode);
    return;
  case Fdeffunc:
    {
      pari_sp ltop=avma;
      long sop,sdat,off;
      long i;
      GEN arg2=listtogen(tree[x].y,Flistarg);
      entree *ep=getfunc(x);
      long loc=y;
      GEN vars=NULL,lvar;
      long nbvar;
      long arity=lg(arg2)-1;
      if (loc>=0)
        while (tree[loc].f==Fseq) loc=tree[loc].x;
      if (ep->valence!=EpNEW && ep->valence!=EpUSER)
      {
        if (ep->valence==EpVAR || ep->valence==EpGVAR)
          pari_err(talker2,"this is a variable",
              tree[n].start,get_origin());
        else
          pari_err(talker2,"cannot redefine GP functions",
              tree[n].start,get_origin());
      }
      if (loc>=0 && tree[loc].f==Ffunction 
          && strcmp(entryname(tree[loc].x),"local")==0)
        vars=listtogen(tree[loc].y,Flistarg);
      nbvar=arity+(vars?lg(vars)-1:0);
      lvar=cgetg(nbvar+1,t_VECSMALL);
      getcodepos(&sop,&sdat,&off);
      for (i=1;i<=arity;i++)
      {
        long a = arg2[lg(arg2)-i];
        long en;
        switch (tree[a].f)
        {
        case Ftag:
          a=tree[a].x;
        case Fentry: /*Fall through*/
          en=(long)getentry(a);
          op_push(OCgetarg,en);
          lvar[i]=en;
          break;
        case Faffect:
          { 
            long lop,ldat,loff;
            getcodepos(&lop,&ldat,&loff);
            compilenode(tree[a].y,Ggen);
            a=tree[a].x;
            if (tree[a].f==Ftag)
              a=tree[a].x;
            op_push(OCpushgen, data_push(getclosure(lop,ldat,loff)));
            en=(long)getentry(a);
            op_push(OCdefaultarg,en);
            lvar[i]=en;
            break;
          }
        default: 
          pari_err(talker2,"invalid function definition",
              tree[a].start,get_origin());
        }
      }
      if (vars)
      {
        for (i=1;i<lg(vars);i++)
        {
          long a=vars[i];
          long en;
          if (tree[a].f==Faffect)
          {
            compilenode(tree[a].y,Ggen);
            a=tree[a].x;
          }
          else
            op_push(OCpushstoi,0);
          if (tree[a].f==Ftag)
            a=tree[a].x;
          en=(long)getentry(a);
          op_push(OCgetarg,en);
          lvar[i+arity]=en;
        }
      }
      if (lg(lvar) > 2)
      { /* check for duplicates */
        GEN x = vecsmall_copy(lvar);
        long k;
        vecsmall_sort(x);
        for (k=x[1],i=2; i<lg(x); k=x[i],i++)
          if (x[i] == k)
            pari_err(talker,"user function %s: variable %s declared twice",
                ep->name, ((entree*)x[i])->name);
      }
      if (y>=0 && tree[y].f!=Fnoarg) compilenode(y,Ggen);
      else compilecast(n,Gvoid,Ggen);
      op_push(OCpushgen, data_push(getclosure(sop,sdat,off)));
      op_push(OCpushgen, data_push(lvar));
      op_push(OCpushgen, data_push(
            strntoGENstr(tree[n].start,tree[n].end-tree[n].start)));
      op_push(OCpushlong, arity);
      op_push(OCdeffunc, (long) ep);
      compilecast(n,Gvoid,mode);
      avma=ltop;
      break;
    }
  case Ftag:
    compilenode(x, mode);
    return;
  default:
    pari_err(bugparier,"compilenode");
  }
}

GEN
gp_closure(long n)
{
  compilenode(n,Ggen);
  return getclosure(0,0,-1); 
}

